# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""this module contains a set of functions to handle inference on astng trees

:version:   $Revision: 1.13 $  
:author:    Sylvain Thenault
:copyright: 2003-2005 LOGILAB S.A. (Paris, FRANCE)
:contact:   http://www.logilab.fr/ -- mailto:python-projects@logilab.org
:copyright: 2003-2005 Sylvain Thenault
:contact:   mailto:thenault@gmail.com
"""

from __future__ import generators

__revision__ = "$Id: inference.py,v 1.13 2006-01-03 14:30:38 syt Exp $"
__doctype__ = "restructuredtext en"

from logilab.common.compat import chain, imap

from logilab.astng import nodes
from logilab.astng._exceptions import InferenceError, NoDefault, \
     NotFoundError, ASTNGBuildingException

#def unpack(stmt):
#    if isinstance(stmt, (nodes.List, nodes.Tuple)):
#        return chain(*imap(unpack, stmt.nodes))
#    return iter( (stmt,) )

def unpack_infer(stmt, path=None):
    """return an iterator on nodes infered by the given statement
    if the infered value is a list or a tuple, recurse on it to
    get values infered by its content
    """
    if isinstance(stmt, (nodes.List, nodes.Tuple)):
        return chain(*imap(unpack_infer, stmt.nodes))
    infered = stmt.infer(path=path).next()
    if infered is stmt:
        return iter( (stmt,) )
    return chain(*imap(unpack_infer, stmt.infer(path=path)))

def path_wrapper(func):
    """return the given infer function wrapped to handle the path"""
    def wrapped(node, name=None, path=None, _func=func, **kwargs):
        """wrapper function handling path"""
        if path is None:
            path = [(node, name)]
        else:
            if (node, name) in path:
                raise StopIteration()
            path.append( (node, name) )
        #print '--'*len(path),_func.__name__[6:], name, path[-1]
        try:
            for res in _func(node, name, path, **kwargs):
                #print '--'*len(path), _func.__name__[6:], '-->', res
                yield res
            if path is not None:
                path.pop()
        except:
            if path is not None:
                path.pop()
            raise
    return wrapped

def _infer_stmts(stmts, name=None, path=None, frame=None):
    """return an iterator on statements infered by each statement in <stmts>
    """
    stmt = None
    one_infered = False
    for stmt in stmts:
        try:
            for infered in stmt.infer(stmt._infer_name(frame, name), path):
                one_infered = True
                yield infered
        except InferenceError:
            ## XXX log error ?
            continue
    if not one_infered:
        raise InferenceError(str(stmt))

# .infert method ##############################################################

def infer_default(self, name=None, path=None):
    """we don't know how to resolve a statement by default"""
    #print 'inference error', self, name, path
    raise InferenceError(self.__class__.__name__)

#infer_default = infer_default
nodes.Node.infer = infer_default


def infer_end(self, name=None, path=None):
    """inference's end for node such as Module, Class, Function, Const...
    """
    yield self

#infer_end = path_wrapper(infer_end)
nodes.Module.infer = nodes.Class.infer = infer_end
nodes.Const.infer = infer_end
nodes.Tuple.infer = infer_end
nodes.Dict.infer = infer_end
nodes.List.infer = infer_end

def infer_function(self, name=None, path=None):
    """infer on Function nodes must be take with care since it
    may be called to infer one of it's argument (in which case <name>
    should be given)
    """
    if name is None:
        yield self
        return
    # Function.argnames can be None in astng (means that we don't have
    # information on argnames)
    if self.argnames is None or not name in self.argnames:
        raise InferenceError()
    try:
        for infered in self.default_value(name).infer(name, path):
            yield infered
    except NoDefault:
        if self.type == 'method' and name == self.argnames[0]:
            yield Instance(self.parent.frame())
        else:
            infer_default(self, name, path)

nodes.Function.infer = path_wrapper(infer_function)
nodes.Lambda.infer = path_wrapper(infer_function)

def infer_name(self, name=None, path=None):
    """infer a Name: use name lookup rules"""
    frame, stmts = self.lookup_name(self.name)
    if not stmts:
        infer_default(self, self.name, path)
    return _infer_stmts(stmts, self.name, path, frame)

nodes.Name.infer = path_wrapper(infer_name)

def infer_assname(self, name=None, path=None):
    """infer a AssName/AssAttr: need to inspect the RHS part of the
    assign node
    """
    stmts = self.assigned_stmts(inf_path=path)
    return _infer_stmts(stmts, self.name, path)
    
nodes.AssName.infer = path_wrapper(infer_assname)

def infer_assattr(self, name=None, path=None):
    """infer a AssName/AssAttr: need to inspect the RHS part of the
    assign node
    """
    stmts = self.assigned_stmts(inf_path=path)
    return _infer_stmts(stmts, self.attrname, path)
    
nodes.AssAttr.infer = path_wrapper(infer_assattr)

def infer_callfunc(self, name=None, path=None):
    """infer a CallFunc node by trying to guess what's the function is
    returning
    """
    one_infered = False
    for callee in self.node.infer(name, path):
        try:
            for infered in callee.infer_call_result(self, path):
                yield infered
                one_infered = True
        except AttributeError:
            ## XXX log error ?
            continue
    if not one_infered:
        raise InferenceError()

nodes.CallFunc.infer = path_wrapper(infer_callfunc)


def infer_getattr(self, name=None, path=None):
    """infer a Getattr node by using getattr on the associated object
    """
    one_infered = False    
    for owner in self.expr.infer(name, path):
        try:
            stmts = owner.getattr(self.attrname)
        except NotFoundError:
            continue
        except AttributeError:
            # XXX List, Dict, Const
            continue
        try:
            for obj in _infer_stmts(stmts, path=path):
                yield obj
                one_infered = True
        except InferenceError:
            continue
    if not one_infered:
        raise InferenceError()
                
nodes.Getattr.infer = path_wrapper(infer_getattr)


def _imported_module_astng(node, modname):
    """return the ast for a module whose name is <modname> imported by <node>
    """
    # handle special case where we are on a package node importing a module
    # using the same name as the package, which may end in an infinite loop
    # on relative imports
    # XXX: no more needed ?
    mymodule = node.root()
    if mymodule.relative_name(modname) == mymodule.name:
        # FIXME: I don't know what to do here...
        raise InferenceError(modname)
    try:
        return mymodule.import_module(modname)
    except (ASTNGBuildingException, SyntaxError):
        raise InferenceError(modname)
        

def infer_import(self, name, path=None, asname=True):
    """self resolve on From / Import nodes return the imported module/object"""
    if name is None:
        infer_default(self, name, path)
    if asname:
        yield _imported_module_astng(self, self.real_name(name))
    else:
        yield _imported_module_astng(self, name)
    
nodes.Import.infer = path_wrapper(infer_import)

def infer_from(self, name, path=None, asname=True):
    """self resolve on From / Import nodes return the imported module/object"""
    if name is None:
        infer_default(self, name, path)
    module = _imported_module_astng(self, self.modname)
    if asname:
        name = self.real_name(name)
    try:
        return _infer_stmts(module.getattr(name), name, path)
    except NotFoundError:
        raise InferenceError(name)

nodes.From.infer = path_wrapper(infer_from)

def infer_global(self, name=None, path=None):
    try:
        yield self.root().resolve(name)
    except NotFoundError:
        raise InferenceError()
nodes.Global.infer = path_wrapper(infer_global)
    

# .infer_call_result method ###################################################

def infer_call_result_function(self, caller, inf_path=None):
    """infer what's a function is returning wen called"""
    if self.is_generator():
        yield Generator(self)
        return
    returns = self.nodes_of_class(nodes.Return, skip_klass=nodes.Function)
    for infered in _infer_stmts(imap(lambda n:n.value, returns), path=inf_path):
        yield infered
    
nodes.Function.infer_call_result = infer_call_result_function

def infer_call_result_class(self, caller, inf_path=None):
    """infer what's a class is returning when called"""
    yield Instance(self)

nodes.Class.infer_call_result = infer_call_result_class


# Assignment related nodes ####################################################
# XXX move into inference.py ?
def assend_assigned_stmts(self, inf_path=None):
    # only infer *real* assignments
    if self.flags == 'OP_DELETE':
        raise InferenceError()
    return self.parent.assigned_stmts(self, inf_path=inf_path)
    
nodes.AssName.assigned_stmts = assend_assigned_stmts
nodes.AssAttr.assigned_stmts = assend_assigned_stmts

def mulass_assigned_stmts(self, node, path=None, inf_path=None):
    if path is None:
        path = []
    node_idx = self.nodes.index(node)
    path.insert(0, node_idx)
    return self.parent.assigned_stmts(self, path, inf_path)
nodes.AssTuple.assigned_stmts = mulass_assigned_stmts
nodes.AssList.assigned_stmts = mulass_assigned_stmts

def assign_assigned_stmts(self, node, path=None, inf_path=None):
    if path is None:
        yield self.expr
        return
    found = False
    for assigned in self.expr.infer(path=inf_path):
        assert path
        for index in path:
            try:
                assigned = assigned.getitem(index)
                assigned = assigned.infer(path=inf_path).next()
            except (AttributeError, IndexError, InferenceError):
                break
        else:
            found = True
            yield assigned
    if not found:
        raise InferenceError()
nodes.Assign.assigned_stmts = assign_assigned_stmts

def tryexcept_assigned_stmts(self, node, path=None, inf_path=None):
    found = False
    for exc_type, exc_obj, body in self.handlers:
        if node is exc_obj:
            for assigned in unpack_infer(exc_type):
                if isinstance(assigned, nodes.Class):
                    assigned = Instance(assigned)
                yield assigned
                found = True
            break
    if not found:
        raise InferenceError()
nodes.TryExcept.assigned_stmts = tryexcept_assigned_stmts
    
def XXX_assigned_stmts(self, node, path=None, inf_path=None):
    raise InferenceError()
nodes.For.assigned_stmts = XXX_assigned_stmts
nodes.ListCompFor.assigned_stmts = XXX_assigned_stmts
nodes.GenExprFor.assigned_stmts = XXX_assigned_stmts


# subscription protocol #######################################################
        
def getitem(self, index):
    return self.nodes[index]
nodes.List.getitem = getitem
nodes.Tuple.getitem = getitem
#Dict.getitem = getitem XXX

# special inference objects ###################################################

class Proxy(object):
    """a simple proxy object"""
    def __init__(self, proxied):
        self._proxied = proxied

    def __getattr__(self, name):
        return getattr(self._proxied, name)

    def infer(self, name=None, path=None):
        yield self

class Instance(Proxy):
    """a special node representing a class instance"""
    def getattr(self, name):
        try:
            return self._proxied.instance_attr(name)
        except NotFoundError:
            return self._proxied.getattr(name)

    def infer_call_result(self, caller, inf_path=None):
        """infer what's a class instance is returning when called"""
        return self._proxied.getattr('__call__').infer_call_result(caller,
                                                                   inf_path)

    def __str__(self):
        return 'Instance of %s' % self._proxied.name
    
class Generator(Proxy): 
    """a special node representing a generator"""
